/*************************************************************************
 * All portions of code are copyright by their respective author/s.
 * Copyright (C) 2007      Bryan Christ <bryan.christ@hp.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------*/

#include <string.h>
#include <form.h>

#include <viper.h>

#include "vwm_scrsaver.h"
#include "vwm.h"
#include "vwm_syssaver.h"

WINDOW*	vwm_fmod_scrsaver(gpointer anything)
{
	VWM			*vwm;
	VWM_MODULE	*scrsaver;
	WINDOW		*window;
	WINDOW		*subwin_menu,*subwin_form;
	MENU			*menu;
	FORM			*form;
	FIELD			**fields;
	gchar			**items;
	GSList		*screen_savers;
	GSList		*node;
	gint			count=0;
	gint			menu_width,menu_height;
	gchar			*buffer;

	vwm=vwm_get_instance();
	screen_savers=vwm_modules_list(VWM_SCREENSAVER);
	if(screen_savers!=NULL) count=g_slist_length(screen_savers);

	items=(gchar**)g_malloc0(count+2);
	items[0]=g_strdup("none");
	node=screen_savers;
	while(node!=NULL)
	{
		scrsaver=(VWM_MODULE*)node->data;
		items[count]=g_strdup(scrsaver->title);
		count--;
		node=node->next;
	}

	if(screen_savers!=NULL) count=g_slist_length(screen_savers);
	items[count+1]=NULL;

	viper_thread_enter();
	menu=viper_menu_create(items);
	g_free(items);
	scale_menu(menu,&menu_height,&menu_width);
	if(menu_width<strlen(" Screensaver ")+8)
		menu_width=strlen(" Screensaver ")+8;

	/*	ugly field buffer	*/
	fields=(FIELD**)g_malloc0(sizeof(FIELD)*2);
	fields[0]=new_field(1,5,1,0,0,0);
	buffer=(gchar*)g_malloc0(10);
	set_field_type(fields[0],TYPE_INTEGER,0,1,90);
	set_field_back(fields[0],VIPER_COLORS(COLOR_WHITE,COLOR_BLACK) | A_BOLD);
	set_field_buffer(fields[0],0,buffer);
	field_opts_off(fields[0],O_AUTOSKIP | O_NULLOK);
	form=new_form(fields);

	window=viper_window_create(" Screensaver ",
		0.5,0.5,menu_width+2,menu_height+9,TRUE);
	subwin_menu=derwin(window,menu_height,menu_width,4,1);
	subwin_form=derwin(window,2,menu_width,menu_height+6,1);
	set_menu_win(menu,subwin_menu);
	set_form_win(form,subwin_form);
	menu_opts_off(menu,O_SHOWDESC);
	set_menu_mark(menu,NULL);
	set_menu_fore(menu,VIPER_COLORS(COLOR_WHITE,COLOR_BLUE) | A_BOLD);
	set_menu_back(menu,VIPER_COLORS(COLOR_BLACK,COLOR_WHITE));
	post_menu(menu);
	post_form(form);

	mvwprintw(window,1,1,"Active Screensaver:");
	mvwprintw(subwin_form,0,0,"Idle Time (min)");
	mvwhline(window,menu_height+5,1,ACS_HLINE,menu_width);
	viper_window_set_userptr(window,(gpointer)menu);
	viper_window_set_key_func(window,vwm_fmod_scrsaver_menu_ON_KEYSTROKE);
	viper_event_set(window,"window-move",
		vwm_fmod_scrsaver_ON_MOVE,(gpointer)subwin_menu);
	viper_event_set(window,"window-close",vwm_fmod_scrsaver_ON_CLOSE,NULL);
	set_menu_userptr(menu,(gpointer)form);

	vwm_fmod_scrsaver_update(window);

	viper_thread_leave();
   g_slist_free(screen_savers);

	return window;
}

gint vwm_fmod_scrsaver_menu_ON_KEYSTROKE(gint32 keystroke,WINDOW *window)
{
	MENU		*menu;
	FORM		*form;
	FIELD		**fields;
	gchar		*buffer;

	menu=(MENU*)viper_window_get_userptr(window);
	form=(FORM*)menu_userptr(menu);
	fields=form_fields(form);

	if(keystroke=='q' || keystroke=='Q')
	{
		viper_thread_enter();
		viper_window_close(window);
		viper_thread_leave();
		return 0;
	}

	if(keystroke==KEY_TAB)
	{
		viper_thread_enter();
      viper_window_set_key_func(window,vwm_fmod_scrsaver_form_ON_KEYSTROKE);
      viper_form_driver(form,REQ_END_FIELD,FORM_CURSOR_BLOCK | FORM_COLORIZE,
         VIPER_COLORS(COLOR_WHITE,COLOR_BLUE) | A_BOLD,
         VIPER_COLORS(COLOR_WHITE,COLOR_BLACK) | A_BOLD,-1); 
		viper_window_redraw(window);
		viper_thread_leave();
		return 0;
	}

	if(keystroke==KEY_CRLF)
	{
		buffer=(gchar*)item_name(current_item(menu));
		if(buffer==NULL) return 0;
      vwm_scrsaver_set(buffer);
		viper_thread_enter();
		viper_window_close(window);
		viper_thread_leave();
      vwm_scrsaver_start();
		return 0;
	}

	if(keystroke==KEY_UP) keystroke=REQ_UP_ITEM;
	if(keystroke==KEY_DOWN) keystroke=REQ_DOWN_ITEM;

	viper_thread_enter();
	menu_driver(menu,keystroke);
	viper_window_redraw(window);
	viper_thread_leave();

	return 0;
}

gint vwm_fmod_scrsaver_form_ON_KEYSTROKE(gint32 keystroke,WINDOW *window)
{
	MENU	*menu;
	FORM	*form;
	FIELD	**fields;
	gchar	*buffer;
   gint  retval;

	menu=(MENU*)viper_window_get_userptr(window);
	form=(FORM*)menu_userptr(menu);
	fields=form_fields(form);

	if(keystroke=='q' || keystroke=='Q')
	{
		viper_thread_enter();
		viper_window_close(window);
		viper_thread_leave();
		return 0;
	}

	if(keystroke==KEY_TAB || keystroke==KEY_CRLF)
	{	
		/* in order for field_buffer() to return the contents of buffer 0,
			it is necessary to call form_driver() and validate the field.	*/
      viper_thread_enter();
		retval=form_driver(form,REQ_VALIDATION);
      viper_thread_leave();
      if(retval==E_INVALID_FIELD) return 0;

		/*	change the screensaver time	*/
		buffer=field_buffer(fields[0],0);
      vwm_scrsaver_timeout_set(atoi(buffer));

      /* clean up the form, change keystroke handler  */
		viper_thread_enter();
      viper_form_normalize(form,VIPER_COLORS(COLOR_WHITE,COLOR_BLACK) | A_BOLD,
         VIPER_COLORS(COLOR_WHITE,COLOR_BLACK) | A_BOLD);
		viper_window_set_key_func(window,vwm_fmod_scrsaver_menu_ON_KEYSTROKE);
      viper_window_redraw(window);
      viper_thread_leave();
      return 0;
	}

	if(keystroke==KEY_DC) keystroke=REQ_DEL_CHAR;
	if(keystroke==KEY_BACKSPACE) keystroke=REQ_DEL_PREV;
	if(keystroke==KEY_HOME) keystroke=REQ_BEG_FIELD;
	if(keystroke==KEY_END) keystroke=REQ_END_FIELD;
	
	viper_thread_enter();
   viper_form_driver(form,keystroke,FORM_CURSOR_BLOCK | FORM_COLORIZE,
      VIPER_COLORS(COLOR_WHITE,COLOR_BLUE) | A_BOLD,
      VIPER_COLORS(COLOR_WHITE,COLOR_BLACK),-1);
	viper_window_redraw(window);
	viper_thread_leave();

	return 0;
}

/*
gint vwm_fmod_scrsaver_ON_ACTIVATE(WINDOW *window,gpointer arg)
{
	vwm_post_help(" [Up/Dn/Tab] Navigate | [Enter] Done | [q] Quit");
	return 0;
}
*/

gint vwm_fmod_scrsaver_ON_MOVE(WINDOW *window,gpointer arg)
{
	viper_thread_enter();
   subwin_move_realign((WINDOW*)arg);
	viper_thread_leave();
	return 0;
}

gint	vwm_fmod_scrsaver_ON_CLOSE(WINDOW *window,gpointer arg)
{
	MENU	*menu;
	FORM	*form;

	menu=(MENU*)viper_window_get_userptr(window);
	form=(FORM*)menu_userptr(menu);

	/*	destroy the menu	*/
	viper_menu_destroy(menu,TRUE);
	/* destroy the form	*/
	viper_form_destroy(form,TRUE);

	return VIPER_EVENT_WINDOW_DESIST;
}

gint vwm_fmod_scrsaver_ON_KEYSTROKE(gint32 keystroke,WINDOW *window)
{
	viper_thread_enter();
	viper_window_close(window);
	viper_thread_leave();

   vwm_scrsaver_reset();
	vwm_scrsaver_state_set(SCRSAVER_STATE_AWOKEN);

	return 1;
}

/* the screensaver tracker */
gint vwm_scrsaver_timer(ps_task_t *task,gpointer anything)
{
	WINDOW					*window;
   VWM_SCRSAVER         *screen_saver;
   VWM_MODULE           *scrsaver_mod;
   gdouble              elapsed;
	static gint				keyring;

   screen_saver=(VWM_SCRSAVER*)anything;
   scrsaver_mod=screen_saver->scrsaver_mod;

   /* check for invalid screensaver configuration. */
   if(scrsaver_mod==NULL || screen_saver->timeout<1)
   {
      /* potential problem.   */
      vwm_scrsaver_stop();
      return PSTHREAD_TERMINATE;
   }

   /* check to see if screensaver timer needs to be restarted. */
   if(screen_saver->state & SCRSAVER_STATE_RESET)
   {
      g_timer_start(screen_saver->timer);
      screen_saver->state &= ~SCRSAVER_STATE_RESET;
      screen_saver->state &= ~SCRSAVER_STATE_RUNNING;
      screen_saver->state |= SCRSAVER_STATE_ACTIVE;
		return PSTHREAD_CONTINUE;
	}

   if(screen_saver->state & SCRSAVER_STATE_AWOKEN)
   {
      vwm_keyring_set(keyring);
      screen_saver->state &= ~SCRSAVER_STATE_AWOKEN;
      return PSTHREAD_CONTINUE;
   }

   /* check to see if timeout has expired and the screensaver needs to run.   */
   elapsed=g_timer_elapsed(screen_saver->timer,NULL);
	if(elapsed>(screen_saver->timeout*60) &&
      !(screen_saver->state & SCRSAVER_STATE_RUNNING))
	{

		viper_thread_enter();
		window=scrsaver_mod->mod_main(scrsaver_mod->anything);
		/*	setting the keyring to -1 disables the hotkeys (at least through
			the vwm_hotkeys_run() facility.	*/
		keyring=vwm_keyring_get();
		vwm_keyring_set(-1);
		/*	override the default key handler	*/
		viper_window_set_key_func(window,vwm_fmod_scrsaver_ON_KEYSTROKE);
		/* make sure the screensaver is eminent	*/
		viper_window_set_state(window,STATE_EMINENT);
		viper_thread_leave();
		screen_saver->state |= SCRSAVER_STATE_RUNNING;
		return PSTHREAD_CONTINUE;
	}

	return PSTHREAD_CONTINUE;
}

/*	updates the message which shows the active screen saver (the currently
	selected screen saver) and the current screen saver idle value.	*/
void vwm_fmod_scrsaver_update(WINDOW *window)
{
	VWM			*vwm;
	MENU			*menu;
	FORM			*form;
	FIELD			**fields;
	gchar			*buffer;

	vwm=vwm_get_instance();
	menu=(MENU*)viper_window_get_userptr(window);
	form=(FORM*)menu_userptr(menu);
	fields=form_fields(form);
	buffer=field_buffer(fields[0],0);

	if(vwm->screen_saver.scrsaver_mod==NULL)
	{
		mvwprintw(window,2,1,"none");
		sprintf(buffer,"0");
	}
	else
	{
 		mvwprintw(window,2,1,"%s",vwm->screen_saver.scrsaver_mod->title);
		sprintf(buffer,"%d",vwm->screen_saver.timeout);
	}
	set_field_buffer(fields[0],0,buffer);

	return;
}

void vwm_scrsaver_start(void)
{
	extern ps_runq_t	*vwm_runq;
   VWM   				*vwm;

   vwm=vwm_get_instance();
   vwm->screen_saver.state |= SCRSAVER_STATE_ACTIVE;
   vwm->screen_saver.state |= SCRSAVER_STATE_RESET;
   if(vwm->screen_saver.timer==NULL)
      vwm->screen_saver.timer=g_timer_new();
   vwm->screen_saver.ps_task=psthread_add(vwm_runq,vwm_scrsaver_timer,
      (gpointer)&vwm->screen_saver);

   return;
}

void vwm_scrsaver_stop(void)
{
	extern ps_runq_t	*vwm_runq;
   VWM   				*vwm;

   vwm=vwm_get_instance();
//   psthread_del(vwm_runq,vwm->screen_saver.ps_task);
   vwm->screen_saver.state=0;
   if(vwm->screen_saver.timer!=NULL) g_timer_stop(vwm->screen_saver.timer);
   return;
}


const gchar* vwm_scrsaver_get(void)
{
   VWM   *vwm;

   vwm=vwm_get_instance();

   if(vwm->screen_saver.scrsaver_mod==NULL) return "none";
   return (const gchar*)vwm->screen_saver.scrsaver_mod->title;
}

void vwm_scrsaver_set(gchar *title)
{
   VWM         *vwm;
   VWM_MODULE  *vwm_module;

   vwm=vwm_get_instance();

   if(strcmp(title,"none")==0)
   {
      vwm->screen_saver.scrsaver_mod=NULL;
      vwm->screen_saver.timeout=0;
      return;
   }   

   vwm_module=vwm_module_find(title);
   if(vwm_module==NULL) return;
   if(strcmp(vwm_module->category,VWM_SCREENSAVER)!=0) return;

   vwm->screen_saver.scrsaver_mod=vwm_module;
   return;
}

void vwm_scrsaver_timeout_set(gint timeout)
{
   VWM   *vwm;

   vwm=vwm_get_instance();

   if(timeout<0) timeout=0;
   vwm->screen_saver.timeout=timeout;

   return;
}   

gint vwm_scrsaver_timeout_get(void)
{
   VWM   *vwm;

   vwm=vwm_get_instance();

   if(vwm->screen_saver.scrsaver_mod==NULL) return 0;
   return vwm->screen_saver.timeout;
}

void vwm_scrsaver_state_set(guint32 state)
{
   VWM   *vwm;

   vwm=vwm_get_instance();   
	vwm->screen_saver.state |= state;

   return;
}
